@@ -17,7 +17,7 @@
  *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
-#define PROGRAM_VERSION "0.1.0"
+#define PROGRAM_VERSION "0.2.0"
 
 #include <arpa/inet.h>
 //#include <sys/types.h>
@@ -42,6 +42,9 @@
 
 #define ADDRESS_CHECK_INTERVAL 1 /* s */
 
+/* level of verbosity */
+int debuglevel;
+
 /* the file to write the PID to */
 char * pidfile;
 
@@ -85,6 +88,12 @@ int ufd_c;
 /* multicast address for sending address changes to */
 struct sockaddr_in multicast_address;
 
+static const char * proto(const char protocol) {
+	if(protocol == UDP) return "udp";
+	if(protocol == TCP) return "tcp";
+	die("proto: invalid protocol");
+}
+
 /* function for allocating memory */
 void allocate_all() {
 	/* allocate memory for leases */
@@ -120,12 +129,12 @@ struct in_addr get_ip_address(const char * ifname) {
 	struct ifreq req;
 	/* len is already checked at command line parsing */
 	//if (strlen(ifname) >= IFNAMSIZ) die("get_ip_address: interface name too long");
-	strncpy(req.ifr_ifrn.ifrn_name, ifname, IFNAMSIZ);
+	strncpy(req.ifr_name, ifname, IFNAMSIZ);
 	int tfd = socket(PF_INET, SOCK_STREAM, 0);
 	int err = ioctl(tfd, SIOCGIFADDR, &req);
 	close(tfd);
 	if (err == 0) {
-		struct sockaddr_in * req_addr = (struct sockaddr_in *) &req.ifr_ifru.ifru_addr;
+		struct sockaddr_in * req_addr = (struct sockaddr_in *) &req.ifr_addr;
 		return (struct in_addr) req_addr->sin_addr;
 	}
 	else {
@@ -180,7 +189,6 @@ void handle_map_request(const int ufd, const struct sockaddr_in * t_addr, const
 	answer_packet.header.op = request_packet->header.op;
 	answer_packet.answer.result = NATPMP_SUCCESS;
 	answer_packet.mapping.private_port = request_packet->mapping.private_port;
-	/* remember: public_port and mapped_port are the same, but differ in name for convience, mapped_port is used where a mapping exists. */
 	answer_packet.mapping.public_port = request_packet->mapping.public_port;
 	answer_packet.mapping.lifetime = request_packet->mapping.lifetime;
 
@@ -189,27 +197,30 @@ void handle_map_request(const int ufd, const struct sockaddr_in * t_addr, const
 
 		if (ntohl(answer_packet.mapping.lifetime) > max_lifetime) {
 			/* lifetime too high, downgrade */
+			if (debuglevel >= 2) printf("Requested lifetime was %u, downgraded to %u\n", ntohl(answer_packet.mapping.lifetime), max_lifetime);
 			answer_packet.mapping.lifetime = htonl(max_lifetime);
 		}
 
 		lease * a = get_lease_by_client_port(client, answer_packet.mapping.private_port);
 		if (a != NULL) {
-			/* lease exists, update expiration time of the requested protocol and answer with mapped port */
+			/* lease exists, update expiration time of the requested protocol and answer with public port */
 			uint32_t new_expires = now + ntohl(answer_packet.mapping.lifetime);
 			if (new_expires == next_lease_expires);
 			else if (new_expires < next_lease_expires) next_lease_expires = new_expires;
 			else if (a->expires[(int) protocol] <= next_lease_expires) update_expires = 1;
 			a->expires[(int) protocol] = new_expires;
-			answer_packet.mapping.mapped_port = a->mapped_port;
+			answer_packet.mapping.public_port = a->public_port;
+			if (debuglevel >= 2) printf("Lease with public %s port %hu for client %s updated\n", proto(protocol), ntohs(answer_packet.mapping.public_port), inet_ntoa(t_addr->sin_addr));
 		}
 		else {
 			/* no lease exists, check for manual mapping */
-			uint16_t mapped_port;
-			int b = get_dnat_rule_by_client_port(protocol, &mapped_port, client, answer_packet.mapping.private_port);
+			uint16_t public_port;
+			int b = get_dnat_rule_by_client_port(protocol, &public_port, client, answer_packet.mapping.private_port);
 			if (b == -1) die("get_dnat_rule_by_client_port returned with error");
 			else if (b == 1) {
-				/* manual mapping exists, answer with mapped port */
-				answer_packet.mapping.mapped_port = mapped_port;
+				/* manual mapping exists, answer with public port */
+				answer_packet.mapping.public_port = public_port;
+				if (debuglevel >= 2) printf("Manual mapping for public %s port %hu for client %s exists\n", proto(protocol), ntohs(answer_packet.mapping.public_port), inet_ntoa(t_addr->sin_addr));
 			}
 			else {
 				/* no lease and no manual mapping exist, find a valid port and create a lease */
@@ -233,12 +244,13 @@ void handle_map_request(const int ufd, const struct sockaddr_in * t_addr, const
 						answer_packet.mapping.public_port = request_packet->mapping.public_port;
 						answer_packet.mapping.lifetime = request_packet->mapping.lifetime;
 						answer_packet.answer.result = NATPMP_OUTOFRESOURCES;
+						if (debuglevel >= 2) printf("No free ports available\n");
 						break;
 					}
 					if (get_lease_by_port(answer_packet.mapping.public_port) == NULL &&
 							is_port_free(answer_packet.mapping.public_port) == 0 &&
-							get_dnat_rule_by_mapped_port(TCP, answer_packet.mapping.public_port, NULL, NULL) == 0 &&
-							get_dnat_rule_by_mapped_port(UDP, answer_packet.mapping.public_port, NULL, NULL) == 0) {
+							get_dnat_rule_by_public_port(TCP, answer_packet.mapping.public_port, NULL, NULL) == 0 &&
+							get_dnat_rule_by_public_port(UDP, answer_packet.mapping.public_port, NULL, NULL) == 0) {
 						/* TODO: acquiring the companion port to a manual mapping can be allowed to the same client */
 						/* these parameters are valid for mapping */
 
@@ -249,7 +261,7 @@ void handle_map_request(const int ufd, const struct sockaddr_in * t_addr, const
 							c.expires[(int) protocol] = now + ntohl(answer_packet.mapping.lifetime);
 							c.client = client;
 							c.private_port = answer_packet.mapping.private_port;
-							c.mapped_port = answer_packet.mapping.mapped_port;
+							c.public_port = answer_packet.mapping.public_port;
 							if (c.expires[(int) protocol] < next_lease_expires) next_lease_expires = c.expires[(int) protocol];
 							add_lease(&c);
 						}
@@ -258,11 +270,13 @@ void handle_map_request(const int ufd, const struct sockaddr_in * t_addr, const
 						{
 							int c = create_dnat_rule(
 									protocol,
-									answer_packet.mapping.mapped_port,
+									answer_packet.mapping.public_port,
 									client,
 									answer_packet.mapping.private_port);
 							if (c == -1) die("create_dnat_rule returned with error");
 						}
+
+						if (debuglevel >= 2) printf("Lease with public %s port %hu for client %s created\n", proto(protocol), ntohs(answer_packet.mapping.public_port), inet_ntoa(t_addr->sin_addr));
 						break;
 					}
 
@@ -281,24 +295,31 @@ void handle_map_request(const int ufd, const struct sockaddr_in * t_addr, const
 		lease * a;
 		int remove_all;
 
-		if (answer_packet.mapping.mapped_port == 0 && answer_packet.mapping.private_port == 0) {
+		if (answer_packet.mapping.public_port == 0 && answer_packet.mapping.private_port == 0) {
 			/* removing all mappings of client (but only for requested protocol) */
 			remove_all = 1;
+			if (debuglevel >= 2) printf("Trying to remove all %s leases for client %s\n", proto(protocol), inet_ntoa(t_addr->sin_addr));
 		}
 		else {
 			/* only removing a single mapping */
 			remove_all = 0;
 			a = get_lease_by_client_port(client, answer_packet.mapping.private_port);
+			if (debuglevel >= 2) printf("Trying to remove lease with public %s port %hu for client %s\n", proto(protocol), ntohs(answer_packet.mapping.public_port), inet_ntoa(t_addr->sin_addr));
 		}
 
 		while (remove_all == 0 || (a = get_next_lease_by_client(client, NULL)) != NULL) {
 			if (a != NULL) {
 				/* destroy mapping */
-				int b = destroy_dnat_rule(protocol, a->mapped_port, client, a->private_port);
+				int b = destroy_dnat_rule(protocol, a->public_port, client, a->private_port);
 				if (b == -1) die("destroy_dnat_rule returned with error");
 				else if (b == 1) {
 					/* mapping may not be destroyed, it's a manual mapping, answer with refused */
 					answer_packet.answer.result = NATPMP_REFUSED;
+					if (debuglevel >= 2) printf("Lease with public %s port %hu for client %s could not be removed\n", proto(protocol), ntohs(a->public_port), inet_ntoa(t_addr->sin_addr));
+				}
+				else
+				{
+					if (debuglevel >= 2) printf("Lease with public %s port %hu for client %s removed\n", proto(protocol), ntohs(a->public_port), inet_ntoa(t_addr->sin_addr));
 				}
 
 				/* update used protocols of lease */
@@ -316,6 +337,7 @@ void handle_map_request(const int ufd, const struct sockaddr_in * t_addr, const
 				else if (b == 1) {
 					/* manual mapping found, answer with refused */
 					answer_packet.answer.result = NATPMP_REFUSED;
+					if (debuglevel >= 2) printf("Lease with public %s port %hu for client %s is mapped manually\n", proto(protocol), answer_packet.mapping.public_port, inet_ntoa(t_addr->sin_addr));
 				}
 			}
 		}
@@ -373,7 +395,7 @@ void fork_to_background() {
 	pid_t child = fork();
 	if (child == -1) p_die("fork");
 	else if (child) {
-		printf("forked into background -- %i\n", child);
+		if (debuglevel >= 1) printf("forked into background -- %i\n", child);
 		if (pidfile != NULL) {
 			FILE * pidfilefd;
 			pidfilefd = fopen(pidfile, "w");
@@ -443,7 +465,7 @@ void update_time() {
 
 //__attribute__ ((noreturn))
 void print_usage(const char * program_name) {
-	fprintf(stderr, "Usage: %s [-b] -i interface -a address [-a address [...]] -- backend-options\n", program_name);
+	fprintf(stderr, "Usage: %s [-v | -q] [-b [-p pidfile]] -i public-interface -a private-address ... [-t max-lifetime] [-l lower-port] [-u upper-port] [[--] backend-options]\n", program_name);
 	exit(EXIT_FAILURE);
 }
 
@@ -471,10 +493,10 @@ void do_version() {
 
 void print_public_ip_address() {
 	if (public_address.s_addr != 0) {
-		printf("IP address of %s: %s\n", public_ifname, inet_ntoa(public_address));
+		if (debuglevel >= 2) printf("IP address of %s: %s\n", public_ifname, inet_ntoa(public_address));
 	}
 	else {
-		printf("IP address of %s: none\n", public_ifname);
+		if (debuglevel >= 2) printf("IP address of %s: none\n", public_ifname);
 	}
 }
 
@@ -486,7 +508,7 @@ void init(int argc, char * argv[]) {
 	port_range_low = 1024; /* ports below 1024 are restricted ports */
 	port_range_high = 65535; /* 65535 is the highest port available */
 
-#define OPTSTRING "Vbp:i:a:t:l:u:"
+#define OPTSTRING "Vvqbp:i:a:t:l:u:"
 	/* parse the command line */
 	{
 		extern char *optarg;
@@ -519,6 +541,7 @@ void init(int argc, char * argv[]) {
 		optind = 0;
 		opterr = -1;
 		pidfile = NULL;
+		debuglevel = 1;
 		memset(public_ifname, 0, sizeof(public_ifname));
 		while ( (opt = getopt(argc, argv, OPTSTRING)) != -1 ) {
 			/* check for nullstrings in the arguments */
@@ -535,6 +558,12 @@ void init(int argc, char * argv[]) {
 
 			/* handle the options */
 			switch (opt) {
+				case 'v': /* verbose */
+					debuglevel = 2;
+					break;
+				case 'q': /* quiet */
+					debuglevel = 0;
+					break;
 				case 'b': /* fork to background */
 					do_fork = -1;
 					break;
@@ -604,9 +633,9 @@ void init(int argc, char * argv[]) {
 		/* catch overflows */
 		if (port_low_offset < port_range_low) port_low_offset = port_range_low;
 
-		printf("Allowed port range: %d..%d, maximal lifetime: %d\n", port_range_low, port_range_high, max_lifetime);
+		if (debuglevel >= 2) printf("Allowed port range: %hu..%hu, maximal lifetime: %u\n", port_range_low, port_range_high, max_lifetime);
 		if (max_lifetime < NATPMP_RECOMMENDED_LIFETIME)
-			fprintf(stderr, "Warning: using maximal lifetime lower than recommended value %d\n", NATPMP_RECOMMENDED_LIFETIME);
+			fprintf(stderr, "Warning: using maximal lifetime lower than recommended value %u\n", NATPMP_RECOMMENDED_LIFETIME);
 
 		public_address = get_ip_address(public_ifname);
 		print_public_ip_address();
@@ -618,7 +647,9 @@ void init(int argc, char * argv[]) {
 			/* prepare data structures for poll */
 			ufd_v[i].events = POLLIN;
 
-			if (laddresses[i].s_addr != 0) printf("Listening on %s\n", inet_ntoa(laddresses[i]));
+			if (laddresses[i].s_addr != 0) {
+				if (debuglevel >= 2) printf("Listening on %s\n", inet_ntoa(laddresses[i]));
+			}
 			else fprintf(stderr, "Warning: Listening on 0.0.0.0 is not a good idea\n");
 		}
 
@@ -651,16 +682,6 @@ void init(int argc, char * argv[]) {
 }
 
 int main(int argc, char * argv[]) {
-#if 0
-	/* test create_dnat_rule() XXX */
-	{
-		struct in_addr address;
-		inet_aton("192.168.1.2", &address);
-		create_dnat_rule(1, htons(81), address.s_addr, htons(80));
-		destroy_dnat_rule(2, htons(81), address.s_addr, htons(80));
-	}
-#endif
-
 	init(argc, argv);
 
 	uint32_t next_address_check = now;
@@ -741,8 +762,12 @@ int main(int argc, char * argv[]) {
 				void destroy_expired(const char protocol) {
 					if (a->expires[(int) protocol] <= now) {
 						a->expires[(int) protocol] = UINT32_MAX;
-						int b = destroy_dnat_rule(protocol, a->mapped_port, a->client, a->private_port);
+						int b = destroy_dnat_rule(protocol, a->public_port, a->client, a->private_port);
 						if (b == -1) die("destroy_dnat_rule returned with error");
+						if (debuglevel >= 2) {
+							struct in_addr client = { a->client };
+							printf("Lease with public %s port %hu for client %s expired and removed\n", proto(protocol), a->public_port, inet_ntoa(client));
+						}
 					}
 				}
 
